# __init__.py - indicates that this directory is a Python package
#
# Copyright 2011 Lincoln de Sousa <lincoln@comum.org>.
# Copyright 2012, 2013, 2014, 2015, 2016 Jeffrey Finkelstein
#           <jeffrey.finkelstein@gmail.com> and contributors.
#
# This file is part of Flask-Restless.
#
# Flask-Restless is distributed under both the GNU Affero General Public
# License version 3 and under the 3-clause BSD license. For more
# information, see LICENSE.AGPL and LICENSE.BSD.
"""Tests that ensure that Flask-Restless meets the requirements of the
JSON API specification.

Each module in this package corresponds to a section of the
specification; for example, the :mod:`test_fetching_data` module
corresponds to the `Fetching Data`_ section of the `JSON API`_
specification.

.. _JSON API: http://jsonapi.org
.. _Fetching Data: http://jsonapi.org/format/#fetching

"""
